!<
!interp
!
!Usage:
!call interp (arg1, arg2 arg3)
!
!Input Parameters:
!
!
!Output Parameters:
!
!
!
!Description:
! This code is a ton of fun.  It contains 8 parts:
! the forward- and backward-running drivers for the preconditioning
! filter, the preconditioning filter itself, the forward- and
! backward-running drivers for the regularization filter, the regularization
! filter itself, and the chain for the forward- and backward-running
! filters for the prec and the reg.
!
! The regularization filter acts along the p_hx axis, first in the
! ``forward'' direction as [0 1 -p] then in the ``backward'' direction
! as [-p 1 0].  The preconditioning filter is the inverse and so chains
! the inverse of the reg. backward filter then the inverse of the reg.
! forward filter.
!
!
!>
!-------------------------------------------------
!
module leaky_mod 
  use oc_adjnull_mod
  use status_mod
  implicit none
  integer, private, save :: ng(3)
  real, private, save :: rho
  contains
  logical function init_leaky(m_s)
    type(sep3d) :: m_s
    init_leaky=.false.
    call from_param("rho",rho,.6)
    ng=(/product(model%n(1:2)),model%n(3),product(model%n(4:5))/)
    init_leaky=.true.
  end function 
  integer function leaky_op(in_tag,out_tag,for_loop,adj)
    logical,            intent(in) :: adj,for_loop
    character(len=128), intent(in) :: in_tag,out_tag
    integer :: nout,ndone,nmax,iout,i1,fwind(4),jwind(4),nwind(4)
    real, allocatable :: model(:),tt(:),d(:)
    integer :: ierr,i2,i,ibdo,ib,ie,ij
    integer :: ibeg,iend,istep
    real :: mult,mult2
    real :: icycle
    allocate(model(ng(1)))
    allocate(data(ng(1)))
    allocate(tt(ng(1)))
    if (ierr.ne.0) then
      call seperr("trouble allocating in leaky")
    end if
    call sseek(in_tag,0,0)
    call sseek(out_tag,0,0)
    if (for_loop) then
      ibeg=1
      iend=ng(2)
      istep=1
    else
      iend=1
      ibeg=ng(2)
      istep=-1
    end if 
    icycle=.0001
    ibdo=0
    do iout=1,ng(3)
      tt=0.
      do i2=ibeg,iend,istep 
        ibdo=        ibdo+1
        mult=(1.-rho)
        mult2=1. 
        if (.not. adj .and. i2.eq.ibeg) then
          mult=1.
        else if ( adj .and. i2.eq.iend) then
          mult2=1./(1.-rho)
        end if
        if (add) then
          call sreed_window(out_tag,3,ng,(/ng(1),1,1/),(/0,i2-1,iout-1&
            &/),(/1,1,1/),4,data)
        else
          data=0
        end if 
        call sreed_window(in_tag,3,ng,(/ng(1),1,1/),(/0,i2-1,iout-1/),&
          &(/1,1,1/),4,model)
        tt=mult2* (rho*tt+model*mult)
        data=        data+tt
        call srite_window(out_tag,3,ng,(/ng(1),1,1/),(/0,i2-1,iout-1/)&
          &,(/1,1,1/),4,data)
      end do
    end do 
    deallocate(tt,model,data)
    leaky_op=0
  end function 
  integer function leaky_bak_op(adj,add,mod_tag,dat_tag,op1,op2,op3,op4&
    &,op5,op6,op7,op8,op9)
    logical,            intent(in) :: adj,add
    character(len=128), intent(in) :: mod_tag,dat_tag
    integer,external,optional :: op1,op2,op3,op4,op5,op6,op7,op8,op9
    if (adj) then
      leaky_bak_op=leaky_op(dat_tag,mod_tag,.true.,.true.)
    else
      leaky_bak_op=leaky_op(mod_tag,dat_tag,.false.,.false.)
    end if
  end function 
  integer function leaky_for_op(adj,add,mod_tag,dat_tag,op1,op2,op3,op4&
    &,op5,op6,op7,op8,op9)
    logical,            intent(in) :: adj,add
    character(len=128), intent(in) :: mod_tag,dat_tag
    integer,external,optional :: op1,op2,op3,op4,op5,op6,op7,op8,op9
    if (adj) then
      leaky_for_op=leaky_op(dat_tag,mod_tag,.false.,.true.)
    else
      leaky_for_op=leaky_op(mod_tag,dat_tag,.true.,.false.)
    end if
  end function 
end module 
