!
!
!Description:
! This code is a ton of fun.  It contains 8 parts:
! the forward- and backward-running drivers for the preconditioning
! filter, the preconditioning filter itself, the forward- and
! backward-running drivers for the regularization filter, the regularization
! filter itself, and the chain for the forward- and backward-running
! filters for the prec and the reg.
!
! The regularization filter acts along the p_hx axis, first in the
! ``forward'' direction as [0 1 -p] then in the ``backward'' direction
! as [-p 1 0].  The preconditioning filter is the inverse and so chains
! the inverse of the reg. backward filter then the inverse of the reg.
! forward filter.
!
! It is based on leaky integration.
!
!
!>
!-------------------------------------------------
!
module geop_prec_mute_mod 
  use oc_adjnull_mod
  use oc_combine_mod
  use mpi_sep
  implicit none
  integer, private, save :: ng(4)
  real, private, save :: rho_in
  integer, private :: maxmem, impi
  character(len=128),private,save :: temp,name,temp2
  logical, private :: mute, mute_beg
  integer, private :: taperx,tapery
  integer, private,allocatable :: intbegmute(:,:,:),intendmute(:,:,:)
  real, private    :: halfpi
  contains
  subroutine geop_prec_init(impi_in,nreg,opx,dpx,mute,mute_modbeg,tag_mod&
    &,maxmem_in)
    integer, intent(in) :: nreg(5)
    real, intent(in) :: opx,dpx
    logical :: mute
    real ,allocatable   :: begmute(:,:,:),endmute(:,:,:)
    integer :: impi_in
    integer, optional   :: maxmem_in
    logical, optional   :: mute_modbeg
    integer  :: iz,ix,iy,ithread,ierr
    integer, external :: sreed
    character(len=128) :: junk
    character(len=128) :: tag_mod
    maxmem=0
    if (present(maxmem_in)) then
      maxmem=maxmem_in
    end if
    mute_beg=.false.
    if (present(mute_modbeg)) then
      mute_beg=mute_modbeg
    end if
    impi=impi_in
    call from_param("rho",rho_in,.95)
    write(0,*)"in geop_prec mute_mod=",mute,nreg
    ng=(/nreg(1),nreg(2),nreg(3),product(nreg(4:5))/)
    name='temp.prec'
    if (impi.eq.0) then
      temp=oc_clone(T,tag_mod,name,maxmem)
    end if
    name='temp2.prec'
    if (impi.eq.0) then
      temp2=oc_clone(T,tag_mod,name,maxmem)
    end if
    call MPI_COMM_RANK(MPI_COMM_WORLD, ithread,ierr)
    call mpi_barrier(mpi_comm_world,ierr)
    if (ithread .eq. 0 .and. mute) then
      if (nreg(4) .ne. 1) then
        call seperr("h_y must equal 1 - CAM")
      end if
      allocate(endmute(nreg(1),nreg(2),nreg(5)),intendmute(nreg(1),nreg&
        &(2),nreg(5)))
      allocate(intbegmute(nreg(1),nreg(2),nreg(5)))
      intbegmute=1
      intendmute=nreg(3)
      call from_param("taperx",taperx,5)
      call from_param("muteend",junk)
      call from_aux("muteend","n1",ierr)
      if (mute_beg) then
        allocate(begmute(nreg(1),nreg(2),nreg(5)))
        if (nreg(1)*nreg(2)*nreg(5)*4 .ne. sreed("mutebeg",begmute,size&
          &(begmute)*4)) then
          call seperr("trouble reading mutebeg")
        end if
      end if
      if (nreg(1)*nreg(2)*nreg(5)*4 .ne. sreed("muteend",endmute,size&
        &(endmute)*4)) then
        call seperr("trouble reading muteend")
      end if
      do iz=1,nreg(5)
        do iy=1,nreg(2)
          do ix=1,nreg(1)
            if (mute_beg) then
              intbegmute(ix,iy,iz)=(begmute(ix,iy,iz)-opx)/dpx+1.5
              if (intbegmute(ix,iy,iz) < 1) then
                intbegmute(ix,iy,iz)=1
              end if
              if (intbegmute(ix,iy,iz) >nreg(3)) then
                intbegmute(ix,iy,iz)=nreg(3)
              end if
            end if
            intendmute(ix,iy,iz)=(endmute(ix,iy,iz)-opx)/dpx+1.5
            if (intendmute(ix,iy,iz) < 1) then
              intendmute(ix,iy,iz)=1
            end if
            if (intendmute(ix,iy,iz) >nreg(3)) then
              intendmute(ix,iy,iz)=nreg(3)
            end if
            if (intendmute(ix,iy,iz) < intbegmute(ix,iy,iz)) then
              intendmute(ix,iy,iz)=intbegmute(ix,iy,iz)
            end if
          end do
        end do
      end do 
      halfpi=1.5707963
      if (mute_beg) then
        deallocate(begmute)
      end if
      deallocate(endmute)
    end if
    call mpi_barrier(mpi_comm_world,ierr)
  end subroutine 
  integer function geop_prec(in_tag,out_tag,for_loop,adj,add)
    logical,            intent(in) :: adj,add,for_loop
    character(len=128), intent(in) :: in_tag,out_tag
    integer :: nout,ndone,nmax,iout,i1,fwind(4),jwind(4),nwind(4)
    complex, allocatable :: model(:,:),tt(:,:),data(:,:)
    integer :: ierr,i2,i,ibdo,ib,ie,ij,ix,iy
    integer :: ibeg,iend,istep
    real :: mult,mult2
    real :: icycle,rho
    integer :: ithread 
    call MPI_COMM_RANK(MPI_COMM_WORLD, ithread,ierr)
    if (ithread .ne. 0) then
      geop_prec=0
      return
    end if
    allocate(model(ng(1),ng(2)))
    allocate(data(ng(1),ng(2)))
    allocate(tt(ng(1),ng(2)))
    if (ierr.ne.0) then
      call seperr("trouble allocating in geop_prec")
    end if
    call sseek(in_tag,0,0)
    call sseek(out_tag,0,0)
    write(0,*)"geop prec, done the sseek",adj
    if (for_loop) then
      ibeg=1
      iend=ng(3)
      istep=1
    else
      iend=1
      ibeg=ng(3)
      istep=-1
    end if 
    icycle=.0001
    ibdo=0
    do iout=1,ng(4)
      tt=0.
      do i2=ibeg,iend,istep 
        ibdo=        ibdo+1
        if (add) then
          call sreed_window(out_tag,4,ng,(/ng(1),ng(2),1,1/),(/0,0,i2-1&
            &,iout-1/),(/1,1,1,1/),8,data)
        else
          data=0
        end if 
        call sreed_window(in_tag,4,ng,(/ng(1),ng(2),1,1/),(/0,0,i2-1&
          &,iout-1/),(/1,1,1,1/),8,model)
        if (.not. mute) then
          rho=rho_in
          mult=(1.-rho)
          mult2=1. 
          if (.not. adj .and. i2.eq.ibeg) then
            mult=1.
          else if ( adj .and. i2.eq.iend) then
            mult2=1./(1.-rho)
          end if
          tt=mult2* (rho*tt+model*mult)
          data=        	data+tt
        else
          do ix=1,ng(1)
            do iy=1,ng(2)
              rho=rho_in
              if (mute_beg) then
                if (i2 < intbegmute(ix,iy,iout)) then
                  rho=0.0
                end if
              end if
              if (i2 > intendmute(ix,iy,iout)) then
                rho=0.0
              end if
              mult=(1.-rho)
              mult2=1. 
              if (.not. adj .and. i2.eq.ibeg) then
                mult=1.
              else if ( adj .and. i2.eq.iend) then
                mult2=1./(1.-rho)
              end if
              tt(ix,iy)=mult2* (rho*tt(ix,iy)+model(ix,iy)*mult)
              data(ix,iy)=        			data(ix,iy)+tt(ix,iy)
            end do
          end do
        end if 
        call srite_window(out_tag,4,ng,(/ng(1),ng(2),1,1/),(/0,0,i2-1&
          &,iout-1/),(/1,1,1,1/),8,data)
      end do
    end do 
    deallocate(tt,model,data)
    geop_prec=0
  end function 
  integer function geop_reg(in_tag,out_tag,for_loop,adj,add)
    logical,            intent(in) :: adj,add,for_loop
    character(len=128), intent(in) :: in_tag,out_tag
    integer :: nout,ndone,nmax,iout,i1,fwind(4),jwind(4),nwind(4)
    complex, allocatable :: model(:),tt(:),data(:)
    integer :: ierr,i2,i,ibdo,ib,ie,ij
    integer :: ibeg,iend,istep
    real :: mult,mult2
    real :: icycle,rho
    allocate(model(ng(1)))
    allocate(data(ng(1)))
    allocate(tt(ng(1)))
    if (ierr.ne.0) then
      call seperr("trouble allocating in geop_prec")
    end if
    call sseek(in_tag,0,0)
    call sseek(out_tag,0,0)
    if (for_loop) then
      ibeg=1
      iend=ng(2)
      istep=1
    else
      iend=1
      ibeg=ng(2)
      istep=-1
    end if 
!need to change this!!!
    icycle=.0001
    ibdo=0
    do iout=1,ng(3)
      tt=0.
      do i2=ibeg,iend,istep 
        ibdo=        ibdo+1
        mult=(1.-rho)
        mult2=1. 
        if (.not. adj .and. i2.eq.ibeg) then
          mult=1.
        else if ( adj .and. i2.eq.iend) then
          mult2=1./(1.-rho)
        end if
        if (add) then
          call sreed_window(out_tag,3,ng,(/ng(1),1,1/),(/0,i2-1,iout-1&
            &/),(/1,1,1/),8,data)
        else
          data=0
        end if 
        call sreed_window(in_tag,3,ng,(/ng(1),1,1/),(/0,i2-1,iout-1/),&
          &(/1,1,1/),8,model)
        tt=mult2* (rho*tt+model*mult)
        data=        data+tt
        call srite_window(out_tag,3,ng,(/ng(1),1,1/),(/0,i2-1,iout-1/)&
          &,(/1,1,1/),8,data)
      end do
    end do 
    deallocate(tt,model,data)
    geop_reg=0
  end function 
  integer function geop_bak_prec(adj,add,mod_tag,dat_tag,op1,op2,op3,op4&
    &,op5,op6,op7,op8,op9)
    logical,            intent(in) :: adj,add
    character(len=128), intent(in) :: mod_tag,dat_tag
    integer,external,optional :: op1,op2,op3,op4,op5,op6,op7,op8,op9
    if (adj) then
      geop_bak_prec=geop_prec(dat_tag,mod_tag,.true.,.true.,add)
    else
      geop_bak_prec=geop_prec(mod_tag,dat_tag,.false.,.false.,add)
    end if
  end function 
  integer function geop_for_prec(adj,add,mod_tag,dat_tag,op1,op2,op3,op4&
    &,op5,op6,op7,op8,op9)
    logical,            intent(in) :: adj,add
    character(len=128), intent(in) :: mod_tag,dat_tag
    integer,external,optional :: op1,op2,op3,op4,op5,op6,op7,op8,op9
    if (adj) then
      geop_for_prec=geop_prec(dat_tag,mod_tag,.false.,.true.,add)
    else
      geop_for_prec=geop_prec(mod_tag,dat_tag,.true.,.false.,add)
    end if
  end function 
  integer function geop_bak_reg(adj,add,mod_tag,dat_tag,op1,op2,op3,op4&
    &,op5,op6,op7,op8,op9)
    logical,            intent(in) :: adj,add
    character(len=128), intent(in) :: mod_tag,dat_tag
    integer,external,optional :: op1,op2,op3,op4,op5,op6,op7,op8,op9
    if (adj) then
      geop_bak_reg=geop_reg(dat_tag,mod_tag,.true.,.true.,add)
    else
      geop_bak_reg=geop_reg(mod_tag,dat_tag,.false.,.false.,add)
    end if
  end function 
  integer function geop_for_reg(adj,add,mod_tag,dat_tag,op1,op2,op3,op4&
    &,op5,op6,op7,op8,op9)
    logical,            intent(in) :: adj,add
    character(len=128), intent(in) :: mod_tag,dat_tag
    integer,external,optional :: op1,op2,op3,op4,op5,op6,op7,op8,op9
    if (adj) then
      geop_for_reg=geop_reg(dat_tag,mod_tag,.false.,.true.,add)
    else
      geop_for_reg=geop_reg(mod_tag,dat_tag,.true.,.false.,add)
    end if
  end function 
  integer function chain_geop_prec(adj,add,mod_tag,dat_tag,op1,op2,op3&
    &,op4,op5,op6,op7,op8,op9)
    logical,            intent(in) :: adj,add
    character(len=128), intent(in) :: mod_tag,dat_tag
    integer,external,optional :: op1,op2,op3,op4,op5,op6,op7,op8,op9
    if (impi .eq. 0) then
      call oc_chain2(geop_bak_prec,geop_for_prec,adj,add,mod_tag,dat_tag&
        &,temp,op1,op2,op3,op4,op5,op6,op7,op8,op9)
    end if
    chain_geop_prec=0
  end function 
  integer function chain_geop_reg(adj,add,mod_tag,dat_tag,op1,op2,op3&
    &,op4,op5,op6,op7,op8,op9)
    logical,            intent(in) :: adj,add
    character(len=128), intent(in) :: mod_tag,dat_tag
    integer,external,optional :: op1,op2,op3,op4,op5,op6,op7,op8,op9
    call oc_chain2(geop_for_reg,geop_bak_reg,adj,add,mod_tag,dat_tag&
      &,temp,op1,op2,op3,op4,op5,op6,op7,op8,op9)
    chain_geop_reg=0
  end function 
end module 
